import builtins
import inspect
from typing import Callable, List, Type, get_type_hints, Any, Dict
import enum

from smarti.class_loader_flags import ClassLoaderFlags
from smarti.exceptions import CyclicDependencyException


class CheckAutowire:
    """A central class to rule them all :)"""
    ANNOTATIONS_MODULE = "smarti.annotations"
    IGNORED_ARGUMENTS = ["self"]

    def __init__(self) -> None:
        self._known_types: List[Type] = []

    def can_autowire(
        self, callable: Callable, flags: ClassLoaderFlags, type_: Type, seen_types: List[Type], kwargs: Dict[str, Any]
    ) -> bool:
        """Checks if a callable can be autowired

        Args:
            callable (Callable): The callable to check
            flags (ClassLoaderFlags): The flags of the classloader
            type_ (Type): The type the callable belongs to
            seen_types (List[Type]): The already instanciated types (CDC)
            kwargs (Dict[str, Any]): All the custom arguments for the function.

        Raises:
            CyclicDependencyException: Is raised if the Callable needs a Type, which needs the type of the callable. e.g. A -> B -> A.

        Returns:
            bool: True if the callable can be autowired, False otherwise.
        """
        hints = get_type_hints(callable)

        signature = inspect.signature(callable)

        for name, param_type in get_type_hints(callable).items():
            if param_type in seen_types and name not in kwargs:
                circle = [
                    t_.__name__ for t_ in seen_types[seen_types.index(param_type):] + [param_type]
                ]
                raise CyclicDependencyException(
                    f"Found cyclic dependencies: {' -> '.join(circle)}")

        problems = {
            k: v
            for k, v in signature.parameters.items()
            if k not in hints
            and k not in CheckAutowire.IGNORED_ARGUMENTS
            and v.kind
            not in [inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD]
            and v.default is inspect.Parameter.empty
        }

        is_autowired = self.is_autowired_or_ignored(type_, flags)

        return not problems and is_autowired

    def can_autowire_type(self, type_: Type) -> bool:
        """Check is a type can be autowired.

        Args:
            type_ (Type): The type to check.

        Returns:
            bool: True if the type can be autowired, False otherwise.
        """
        type_to_check = type_

        if type.__module__.endswith(CheckAutowire.ANNOTATIONS_MODULE):
            type_to_check = type_.__bases__[0]

        type_str = type_to_check.__name__

        return (
            type_str not in [name for name in dir(
                builtins) if name[0].islower()]
            and not inspect.isabstract(type_to_check)
            and not issubclass(type_to_check, enum.Enum)
        )

    def is_autowired_or_ignored(self, class_: Type, flags: ClassLoaderFlags) -> bool:
        """Determines if the class is autowired or if the ignore flag is used.

        Args:
            class_ (Type): The class to check.
            flags (ClassLoaderFlags): The class loader flags.

        Returns:
            bool: True if the class is autowired or if autowireing is not necessary, else False.
        """
        is_autowired = self.is_autowired(class_)
        force_all_autowired = bool(
            flags & ClassLoaderFlags.ALL_DEPENDENCIES_AUTOWIRED)
        ignore = bool(flags & ClassLoaderFlags.IGNORE_POSSIBLE_THREAD_ERRORS)

        return is_autowired or not force_all_autowired or ignore

    def is_autowired(self, class_: Type) -> bool:
        """Check if a class is autpwired.

        Args:
            class_ (Type): The class to check.

        Returns:
            bool: True if the class is autowired, else False.
        """
        return class_ in self._known_types

    def is_thread_safe(self, flags: ClassLoaderFlags) -> bool:
        """Determines if thread-safe mode is set.

        Args:
            flags (ClassLoaderFlags): The class loader flags.

        Returns:
            bool: True if thread-safe mode is active, else False.
        """
        return bool(
            flags & ClassLoaderFlags.ALL_DEPENDENCIES_AUTOWIRED
            or flags & ClassLoaderFlags.IGNORE_POSSIBLE_THREAD_ERRORS
        )



from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar
import inspect
from threading import Lock
import pickle
from smarti import constants as cst

T = TypeVar("T")


class InstanceStorage:
    """This class handles all the singleton related magic.
    """
    IGNORED_ARGUMENTS = [cst.ALREADY_SEEN_TYPES]

    def __init__(self) -> None:
        self._storage: Dict[Tuple, Any] = {}
        self._storage_lock = Lock()

    def get_instance(
        self, type_: Type[T], arguments: List, kwargs: Optional[Dict] = None
    ) -> Optional[T]:
        """Get an existing instance or None from the storage.

        Args:
            type_ (Type[T]): The type of the instance.
            arguments (List): The arguments of the instance,
            kwargs (Optional[Dict], optional): The kwargs of the instance. Defaults to None.

        Raises:
            RuntimeError: If the module of the class cannot be found.

        Returns:
            Optional[T]: The instance of the type with the given arguments. None if there is no such instance.
        """
        classname = type_.__name__

        module = inspect.getmodule(type_)
        if module is None:
            raise RuntimeError(f"Could not get module of type {type_}")

        key = self._generate_key(
            module.__name__, classname, type_, arguments, kwargs)

        self._storage_lock.acquire()
        data = self._storage.get(key, None)
        self._storage_lock.release()

        return data

    def add_or_get(
        self, type_: Type[T], instance: T, arguments: List, kwargs: Optional[Dict] = None
    ) -> T:
        """Adds a new instance or gets the equal instance from the sotrage.

        Args:
            type_ (Type[T]): The class of the instance.
            instance (T): The instance itself.
            arguments (List): The arguments of the instance.
            kwargs (Optional[Dict], optional): The kwargs for the instance. Defaults to None.

        Raises:
            RuntimeError: If the module could not be found.

        Returns:
            T: The instance.
        """
        classname = type_.__name__
        module = inspect.getmodule(type_)
        if module is None:
            raise RuntimeError(f"Could not get module of type {type_}")

        key = self._generate_key(
            module.__name__, classname, type_, arguments, kwargs)

        self._storage_lock.acquire()
        data = self._storage.get(key, None)
        if data:
            self._storage_lock.release()
            return data

        self._storage[key] = instance
        self._storage_lock.release()

        return instance

    def _generate_key(
        self,
        module: str,
        classname: str,
        type_: Type,
        arguments: List,
        kwargs: Optional[Dict],
    ) -> Tuple:
        """Generates the key for the instance storage.

        Args:
            module (str): The module of the instance.
            classname (str): The classname of the instance.
            type_ (Type): The type of the instance,
            arguments (List): The arguments of the instance.
            kwargs (Optional[Dict]): The kwargs of the instance.

        Returns:
            Tuple: The key for the local instance storage.
        """
        sig = inspect.signature(type_.__init__)
        params = [i for i in sig.parameters.items() if i[0] != 'self']
        args = []

        for i, (name, param) in enumerate(params):
            if param.kind == inspect.Parameter.VAR_POSITIONAL:
                try:
                    args.append((name, pickle.dumps(arguments[i:])))
                except TypeError:
                    tmp_args: List[Any] = []
                    for k in kwargs:  # type: ignore
                        try:
                            tmp_args.append(pickle.dumps(k))
                        except TypeError:
                            tmp_args.append(k)
                    args.append(tuple(tmp_args))
                continue
            elif param.kind == inspect.Parameter.VAR_KEYWORD:
                continue

            args.append((name, pickle.dumps(arguments[i])))

        try:
            kw_arg_hashable = pickle.dumps(kwargs)
        except TypeError:
            tmp_args: List[Any] = []
            for k, v in kwargs.items():  # type: ignore
                try:
                    tmp_args.append(pickle.dumps((k, v)))
                except TypeError:
                    tmp_args.append((k, v))

            kw_arg_hashable = tuple(tmp_args)  # type: ignore

        key = tuple([f"{module}.{classname}", *args, kw_arg_hashable])

        return key



from typing import Optional, Type, TypeVar
import smarti.class_loader as cl
import smarti.constants as cst

GLOBAL_CLASSLOADER = cl.ClassLoader()
T = TypeVar("T")


def autowired(
    class_: Type[T] = None,
    as_singleton: bool = True,
    class_loader: Optional[cl.ClassLoader] = None,
    **kwargs
):
    """The main decorator of this package. It allows to autowire classes by decorating them. It also supports singletons and custom class loader!
    If positional arguments are present, the whole autowireing is skipped to enable backwards compatability.

    Args:
        class_ (Type[T], optional): The class, typically inserted by python itself using the decorator syntax. Defaults to None.
        as_singleton (bool, optional): True if this class should be loaded as a singleton, False otherwise. Defaults to True.
        class_loader (Optional[cl.ClassLoader], optional): The custom class loader. If None smarti.decorator.GLOBAL_CLASSLOADER will be used. Defaults to None.
    """
    def decorator(decorated_class: Type[T]):
        used_class_loader = GLOBAL_CLASSLOADER if class_loader is None else class_loader
        annotation_args = kwargs

        def __new__(cls, *args, **kwargs) -> T:
            if as_singleton:
                existing_instance = used_class_loader._instance_storage.get_instance(
                    decorated_class, list(args), kwargs
                )
                if existing_instance:
                    return existing_instance

            original_new = getattr(decorated_class, cst.UNMODIFIED_NEW)
            return original_new(cls)

        def __init__(self, *args, **kwargs) -> None:
            if args:
                original_init = getattr(decorated_class, cst.UNMODIFIED_INIT)
                original_init(self, *args)
                return

            existing_instance = None
            if as_singleton:
                existing_instance = used_class_loader._instance_storage.get_instance(
                    decorated_class, [], kwargs
                )

            if not existing_instance:
                original_init = getattr(decorated_class, cst.UNMODIFIED_INIT)
                seen_types = kwargs.get(cst.ALREADY_SEEN_TYPES, [])
                used_class_loader.autowire_function(
                    decorated_class,
                    original_init,
                    self,
                    as_singleton,
                    seen_types,
                    **{**annotation_args, **kwargs}
                )

                used_class_loader._instance_storage.add_or_get(
                    decorated_class, self, [], kwargs
                )

        setattr(decorated_class, cst.UNMODIFIED_INIT, decorated_class.__init__)
        setattr(decorated_class, cst.UNMODIFIED_NEW, decorated_class.__new__)
        decorated_class.__init__ = __init__  # type: ignore
        decorated_class.__new__ = __new__  # type: ignore

        dont_add = (
            kwargs[cst.DONT_ADD_TO_KNOWN] if cst.DONT_ADD_TO_KNOWN in kwargs else False
        )

        if not dont_add:
            used_class_loader._check_autowire._known_types.append(
                decorated_class)

        return decorated_class

    if class_ is None:
        return decorator
    else:
        return decorator(class_)



from enum import Flag
import enum


class ClassLoaderFlags(Flag):
    """Flags for the smarti.ClassLoader"""
    ALL_DEPENDENCIES_AUTOWIRED = enum.auto()
    IGNORE_POSSIBLE_THREAD_ERRORS = enum.auto()

    NO_FLAGS = enum.auto()



from distutils.core import setup
from smarti import __version__ as smarti_version

try:
    with open("README.md", 'r') as file:
        content = file.read()
except FileNotFoundError:
    content = "A smart and lightweight dependency injector, supporting singletons and different containers."

setup(
    name="smarti",
    version=smarti_version,
    description="A smart and lightweight dependency injector",
    author="optzGuitar",
    url="https://github.com/optzGuitar/smarti",
    packages=["smarti"],
    long_description=content,
    long_description_content_type='text/markdown'
)



import inspect
import smarti as sti
import smarti.constants as cst
from typing import Any, Callable, Dict, List, Optional, Type, get_type_hints, TypeVar

from smarti.check_autowire import CheckAutowire
from smarti.class_loader_flags import ClassLoaderFlags
from smarti.instance_storage import InstanceStorage

T = TypeVar('T')


class ClassLoader:
    """This classloader is responsible for instanciating the new instances. It automatically detects if a class is autowired or not and loads it correspondingly."""

    def __init__(self, flags: ClassLoaderFlags = ClassLoaderFlags.NO_FLAGS) -> None:
        self._check_autowire = CheckAutowire()
        self._instance_storage = InstanceStorage()

        self.set_flags(flags)

    def set_flags(self, flags: ClassLoaderFlags):
        """Updates the flags to this ClassLoader.

        Args:
            flags (ClassLoaderFlags): The new flags.
        """
        self._flags = flags

    def manually_add_instance(self, type_: Type[T], instance: T, arguments: List[Any], kwargs: Optional[Dict[str, Any]] = None) -> T:
        """Manually adds an instance to the instance storage. This has only an effect if singletons are used!

        Args:
            type_ (Type[T]): The type of the instance
            instance (T): The instance itself.
            arguments (List[Any]): The arguments needed to instantiate the instance.
            kwargs (Optional[Dict[str, Any]], optional): Custom KWargs of the instance. Defaults to None.

        Returns:
            T: The instance or an equal already saved instance.
        """
        return self._instance_storage.add_or_get(type_, instance, arguments, kwargs)

    def autowire_function(
        self,
        type_: Type,
        function: Callable,
        self_arg: Any,
        as_singleton: bool,
        seen_types: List[Type],
        **kwargs,
    ):
        """Autowired a callable.

        Args:
            type_ (Type): The type the callable belongs to.
            function (Callable): The callable itself.
            self_arg (Any): The self-arg of the callable.
            as_singleton (bool): True if singletons should be used, False otherwise.
            seen_types (List[Type]): The already instanciated types of this depencency chain.

        Raises:
            RuntimeError: If the function cannot be autowired.
            TypeError: If a type cannot be autowired.
            CyclicDependencyException: If there exists a cyclic dependency between types.
        """
        args = [self_arg]

        sign = inspect.signature(function)
        if not self._check_autowire.can_autowire(function, self._flags, type_, seen_types, kwargs):
            raise RuntimeError(f"Cannot Autowire function {function}")

        for name, arg_type in get_type_hints(function).items():
            if name == "self" or name == "return":
                continue

            if name in kwargs:
                args.append(kwargs[name])
                continue

            if sign.parameters[name].default is not inspect.Parameter.empty:
                continue

            if not self._check_autowire.can_autowire_type(arg_type):
                raise TypeError(
                    f"Cannot Autowire {name}: {arg_type} of {function}")

            args.append(self._instantiate_class(
                arg_type, name, kwargs, as_singleton, [*seen_types, arg_type]))

        function(*args)

    def _instantiate_class(
        self, type: Type[T], name: str, kwargs: Dict[str, Any], as_singleton: bool, seen_types: List[Type]
    ) -> T:
        """Instantiate a class.

        Args:
            type (Type[T]): The type to instantiate
            name (str): Argument name for the kwargs
            kwargs (Dict[str, Any]): The kwargs.
            as_singleton (bool): True if singletons should be used, False otherwise.
            seen_types (List[Type]): The already instanciated types of this depencency chain.

        Returns:
            T: The new or stored instance of the type.
        """
        class_ = self._load_class_type(type)

        if self._check_autowire.is_autowired(class_):
            custom_args = self._get_kwargs_for_argument(name, kwargs)
            instance = self._create_instance(class_, custom_args, seen_types)
        else:
            instance = self._recursive_instantiate(
                class_, as_singleton, name, kwargs, seen_types)

        return instance

    def _recursive_instantiate(
        self, class_: Type[T], as_singleton: bool, name: str, kwargs: Dict[str, Any], seen_types: List[Type],
    ) -> T:
        """Recursively instantiates a class.

        Args:
            class_ (Type[T]): The class to instantiate
            as_singleton (bool): If singletons should be used.
            name (str): The name of the argument.
            kwargs (Dict[str, Any]): The kwargs.
            seen_types (List[Type]): The already instanciated types of this depencency chain.

        Raises:
            RuntimeError: If thread-safety is flagged and one of the classes of the dependency chain is not autowired.

        Returns:
            T: The instance.
        """
        custom_args = self._get_kwargs_for_argument(name, kwargs)

        # TODO: check if thread safety is an option; due to new decorator may be not possible anymore
        if self._check_autowire.is_thread_safe(self._flags):
            raise RuntimeError(
                f"Cannot create unwired class in {ClassLoaderFlags.ALL_DEPENDENCIES_AUTOWIRED} mode"
            )

        class_ = sti.autowired(
            class_, as_singleton, self, **{cst.DONT_ADD_TO_KNOWN: True}
        )
        instance = self._create_instance(class_, custom_args, seen_types)

        class_.__init__ = getattr(class_, cst.UNMODIFIED_INIT)  # type: ignore
        class_.__new__ = getattr(class_, cst.UNMODIFIED_NEW)  # type: ignore

        return instance

    def _create_instance(self, type_: Type[T], custom_args: Dict[str, Any], seen_types: List[Type]) -> T:
        try:
            return type_(
                **{**custom_args, cst.ALREADY_SEEN_TYPES: seen_types})  # type: ignore
        except TypeError as e:
            raise TypeError(f"Cannot create {type_} with custom args {custom_args} and dependency chain {seen_types}") from e

    def _load_class_type(self, type: Type) -> Any:
        """Gets the type of a type.

        Args:
            type (Type): The type to get the type from.

        Returns:
            Any: Sonething
        """
        module = inspect.getmodule(type)

        return getattr(module, type.__name__)

    def _get_kwargs_for_argument(
        self, name: str, kwargs: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Gets the custom args for the given argument.

        Args:
            name (str): The argument name.
            kwargs (Dict[str, Any]): All the kwargs.

        Returns:
            Dict[str, Any]: The kwargs for the given argument.
        """
        kwa = kwargs.get(f"{name}{cst.KWARGS_VALUE}", None)

        if not kwa:
            return {}

        return kwa
